import datetime
import functools
from http import HTTPStatus

from posixpath import join as posixjoin
from typing import Optional, List

from sermonaudio import API, APIException, _session
from sermonaudio.models import Sermon, SermonEventType
from sermonaudio.utils import update_kwargs_for_key

URL_PATH_BROADCASTER = 'broadcaster'


class BroadcasterAPIError(APIException):  # pragma: no cover
    pass


def parse_sermon(response) -> Sermon:  # pragma: no cover
    if response.ok:
        return Sermon.parse(response.json())
    else:
        raise BroadcasterAPIError(response.json())


class Broadcaster(API):  # pragma: no cover. Unfortunately we can't test this w/o a test account...
    @classmethod
    def create_or_update_sermon(cls,
                                sermon_id: Optional[str],
                                accept_copyright: bool,
                                full_title: str,
                                speaker_name: str,
                                preach_date: datetime.date,
                                publish_date: Optional[datetime.date],
                                event_type: SermonEventType,
                                display_title: Optional[str],
                                subtitle: Optional[str],
                                bible_text: Optional[str],  # TODO: structure this data
                                more_info_text: Optional[str],
                                language_code: str,
                                keywords: Optional[List[str]]) -> Sermon:
        """Creates a new sermon record or updates an existing one (media uploaded separately).

        :param sermon_id: The sermon ID, if you are updating an existing sermon. Otherwise, None.
        :param accept_copyright: A boolean value indicating that you agree that you are allowed to upload this content.
        :param full_title: The full sermon title.
        :param speaker_name: The speaker name (please be consistent; speakers will be created if they don't exist).
        :param preach_date: The date the sermon was preached.
        :param publish_date: The date that the sermon should be visible on the site. You must submit this in order
        for the sermon to be visible to the public. Regrettably, we US Eastern time rather than UTC for
        historical reasons.
        :param event_type: The type of event that this sermon was preached at.
        :param display_title: An alternate, shorter version of the title to be displayed in ID3v1 tags and parts of
        the legacy site (30 char limit)
        :param subtitle: The subtitle of the sermon (or the series, if multiple sermons use the same subtitle; this
        will likely be reworked in the future)
        :param bible_text: The scripture passage(s) that this sermon was derived from
        :param more_info_text: Additional info about the sermon that you wish to share
        :param language_code: The ISO 639 language code for the sermon.
        :param keywords: A list of keywords for this sermon
        :return: The created sermon, if successful, or None
        """
        params = {
            'acceptCopyright': accept_copyright,
            'fullTitle': full_title,
            'speakerName': speaker_name,
            'preachDate': preach_date.isoformat(),
            'publishDate': publish_date.isoformat() if publish_date else None,
            'eventType': event_type.value,
            'displayTitle': display_title,
            'subtitle': subtitle,
            'bibleText': bible_text,
            'moreInfoText': more_info_text,
            'languageCode': language_code,
            'keywords': keywords.join(' ') if keywords else None,
        }

        if sermon_id:
            path = posixjoin('node', 'sermons', sermon_id)
            method = 'put'
        else:
            path = posixjoin('node', 'sermons')
            method = 'post'

        return cls._request(path,
                            method,
                            params=params,
                            parse_func=parse_sermon)

    @classmethod
    def _upload_media(cls,
                      upload_type: str,
                      sermon_id: str,
                      path: str,
                      **kwargs):  # pragma: no cover
        """Uploads media for a sermon.
        :param upload_type: The type of media to upload.
        :param sermon_id: The Sermon ID you are uploading media for.
        :param path: The path to the file on disk.
        :param kwargs: Additional arguments to pass to the underlying post method.
        :return: No return value. Throws a BroadcasterAPIError with details on failure.
        """
        update_kwargs_for_key(kwargs, 'params', {
            'uploadType': upload_type,
            'sermonID': sermon_id,
        })

        response = cls.post('media', **kwargs)

        if response.status_code != HTTPStatus.CREATED:
            raise BroadcasterAPIError({
                'error': 'Unable to create media upload',
                'json': response.json(),
            })

        upload_url = response.json()['uploadURL']

        with open(path, 'rb') as fp:
            response = _session.post(upload_url,
                                     data=fp,
                                     stream=True)
            if not response.ok:
                raise BroadcasterAPIError({
                    'error': f'Received unexpected HTTP status code {response.status_code} when uploading data.',
                    'response': response.content
                })

    upload_audio = functools.partialmethod(_upload_media, 'original-audio')
    upload_video = functools.partialmethod(_upload_media, 'original-video')



"""Misc utility functions"""


def update_kwargs_for_key(kwargs: dict, key: str, values: dict):
    """Updates the passed in dictionary for the given key (using the dictionary `update` method of merging).

    If no value exists for the key, the supplied values dict is assigned directly.

    :param kwargs: The kwargs dictionary to modify in-place.
    :param key: The key to set (e.g. kwargs[key]).
    :param values: The values to update (e.g. kwargs[key].update(values)).
    """
    if key in kwargs:
        kwargs[key].update(values)
    else:
        kwargs[key] = values



"""API data structures"""
import datetime
import dateutil.parser
import enum
import pytz

_model_override = {}


class MultipleOverrideError(Exception):
    pass


class Model:
    """Base model class.

    Presently exists mostly for type hinting and providing a basic interface since Python lacks interfaces/protocols.
    """
    def __init__(self, obj: dict):
        super().__init__()

        self.__obj = obj

    @classmethod
    def parse(cls, obj: dict):
        return _model_override.get(cls, cls)(obj)


def override_model(model):
    def wrapper(new_model):
        if model in _model_override:
            raise MultipleOverrideError(f'Multiple overrides for {model}. This is almost certainly an error.'
                                        'If you ABSOLUTELY have a good reason for doing this, you should'
                                        'instead override the existing subclass of this model and let the'
                                        'resolution chain work deterministically. Any other use is assuredly'
                                        'going to land you in a world of pain.')

        _model_override[model] = new_model

        return new_model

    return wrapper


def isinstance_or_none(obj, t):
    return obj is None or isinstance(obj, t)


class Node(Model):
    """The base node object, which encapsulates all node API responses"""

    def __init__(self, obj: dict):
        super().__init__(obj)

        self.node_type = obj['nodeType']
        assert isinstance(self.node_type, str)

        self.node_display_name = obj['nodeDisplayName']
        assert isinstance(self.node_display_name, str)

        self.results = obj['results']
        assert (isinstance(self.results, list) or
                isinstance(self.results, dict) or
                self.results is None)

        self.total_count = obj['totalCount']
        assert isinstance_or_none(self.total_count, int)

        self.next = obj['next']
        assert isinstance_or_none(self.next, str)


class Broadcaster(Model):
    def __init__(self, obj: dict):
        super().__init__(obj)

        self.broadcaster_id = obj['broadcasterID']
        assert isinstance(self.broadcaster_id, str)

        self.service_times_are_preformatted = obj['serviceTimesArePreformatted']
        assert isinstance_or_none(self.service_times_are_preformatted, bool)

        self.service_times = obj['serviceTimes']
        assert isinstance_or_none(self.service_times, str)

        self.denomination = obj['denomination']
        assert isinstance_or_none(self.denomination, str)

        self.address = obj['address']
        assert isinstance_or_none(self.address, str)

        self.display_name = obj['displayName']
        assert isinstance(self.display_name, str)

        self.location = obj['location']
        assert isinstance(self.location, str)

        self.latitude = obj['latitude']
        assert isinstance_or_none(self.latitude, float)

        self.longitude = obj['longitude']
        assert isinstance_or_none(self.longitude, float)

        self.image_url = obj['imageURL']
        assert isinstance(self.image_url, str)

        self.album_art_url_format = obj['albumArtURL']
        assert isinstance(self.album_art_url_format, str)

        self.minister = obj['minister']
        assert isinstance_or_none(self.minister, str)

        self.phone = obj['phone']
        assert isinstance_or_none(self.phone, str)

        self.home_page_url = obj['homePageURL']
        assert isinstance_or_none(self.home_page_url, str)

        self.bible_version = obj['bibleVersion']
        assert isinstance_or_none(self.bible_version, str)

        self.facebook_username = obj['facebookUsername']
        assert isinstance_or_none(self.facebook_username, str)

        self.twitter_username = obj['twitterUsername']
        assert isinstance_or_none(self.twitter_username, str)

        self.about_us = obj['aboutUs']
        assert isinstance_or_none(self.about_us, str)

        self.webcast_in_progress = obj['webcastInProgress']
        assert isinstance(self.webcast_in_progress, bool)

        self.vacant_pulpit = obj['vacantPulpit']
        assert isinstance(self.vacant_pulpit, bool)

    def get_album_art_url(self, size: int):  # pragma: no cover
        """Returns a URL for the square album art with a with and height equal to the provided size argument"""
        return self.album_art_url_format.replace('{size}', str(size))

    def __str__(self):  # pragma: no cover
        return f'<Broadcaster {self.broadcaster_id} "{self.display_name}">'


class Speaker(Model):
    def __init__(self, obj: dict):
        super().__init__(obj)

        self.display_name = obj['displayName']
        assert isinstance(self.display_name, str)

        self.sort_name = obj.get('sortName')
        assert isinstance(self.sort_name, str)

        self.bio = obj['bio']
        assert isinstance_or_none(self.bio, str)

        self.portrait_url = obj['portraitURL']
        assert isinstance(self.portrait_url, str)

        self.rounded_thumbnail_image_url = obj['roundedThumbnailImageURL']
        assert isinstance(self.rounded_thumbnail_image_url, str)

        self.album_art_url_format = obj['albumArtURL']
        assert isinstance(self.album_art_url_format, str)

    def get_album_art_url(self, size: int):  # pragma: no cover
        """Returns a URL for the square album art with a with and height equal to the provided size argument"""
        return self.album_art_url_format.replace('{size}', str(size))


@enum.unique
class MediaType(enum.Enum):
    MP3 = 'mp3'
    AAC = 'aac'

    MP4 = 'mp4'

    PDF = 'pdf'
    WORD = 'doc'
    TRANSCRIPT = 'transcript'

    JPEG = 'jpg'

    ORIGINAL_AUDIO = 'orig-audio'
    ORIGINAL_VIDEO = 'orig-video'


class Media(Model):
    def __init__(self, obj: dict):
        super().__init__(obj)

        self.media_type = MediaType(obj['mediaType'])
        assert isinstance(self.media_type, MediaType)

        self.is_live = obj['live']
        assert isinstance(self.is_live, bool)

        self.stream_url = obj['streamURL']
        assert isinstance_or_none(self.stream_url, str)

        self.download_url = obj.get('downloadURL')
        assert isinstance_or_none(self.download_url, str)

        self.bitrate = obj['bitrate']
        assert isinstance_or_none(self.bitrate, int)

        self.duration = obj['duration']
        assert isinstance_or_none(self.duration, int)

        self.audio_codec = obj['audioCodec']
        assert isinstance_or_none(self.audio_codec, str)

        self.video_codec = obj['videoCodec']
        assert isinstance_or_none(self.video_codec, str)

        self.thumbnail_image_url = obj['thumbnailImageURL']
        assert isinstance_or_none(self.thumbnail_image_url, str)

        self.raw_url = obj.get('rawURL')
        assert isinstance_or_none(self.raw_url, str)


class MediaSet(Model):
    def __init__(self, obj: dict):
        super().__init__(obj)

        self.audio = [Media.parse(rec) for rec in obj['audio']]
        self.video = [Media.parse(rec) for rec in obj['video']]
        self.text = [Media.parse(rec) for rec in obj['text']]


class Sermon(Model):
    def __init__(self, obj: dict):
        super().__init__(obj)

        self.sermon_id = obj['sermonID']
        assert isinstance(self.sermon_id, str)

        self.broadcaster = Broadcaster.parse(obj['broadcaster'])
        self.speaker = Speaker.parse(obj['speaker'])

        self.full_title = obj['fullTitle']
        assert isinstance(self.full_title, str)

        self.display_title = obj['displayTitle']
        assert isinstance(self.display_title, str)

        self.subtitle = obj['subtitle']
        assert isinstance_or_none(self.subtitle, str)

        self.preach_date = datetime.datetime.strptime(obj['preachDate'], '%Y-%m-%d').date()
        assert isinstance(self.preach_date, datetime.date)

        try:
            self.staff_pick_date = datetime.datetime.strptime(obj['pickDate'], '%Y-%m-%d').date()
            assert isinstance_or_none(self.staff_pick_date, datetime.date)
        except (TypeError, ValueError):
            self.staff_pick_date = None

        try:
            self.publish_date = datetime.datetime.strptime(obj['publishDate'], '%Y-%m-%d').date()
            assert isinstance_or_none(self.publish_date, datetime.date)
        except (TypeError, ValueError):  # pragma: no cover
            self.publish_date = None

        self.language_code = obj['languageCode']
        assert isinstance(self.language_code, str)

        self.bible_text = obj['bibleText']
        assert isinstance_or_none(self.bible_text, str)

        self.more_info_text = obj['moreInfoText']
        assert isinstance_or_none(self.more_info_text, str)

        self.event_type = obj['eventType']
        assert isinstance(self.event_type, str)

        self.download_count = obj['downloadCount']
        assert isinstance(self.download_count, int)

        self.external_link = obj['externalLink']
        assert isinstance_or_none(self.external_link, str)

        self.keywords = obj.get('keywords', [])

        self.media = MediaSet.parse(obj['media'])

    def __str__(self):  # pragma: no cover
        return f'<Speaker {self.speaker.display_name} - {self.display_title}>'


class Webcast(Model):
    def __init__(self, obj: dict):
        super().__init__(obj)

        self.display_name = obj['displayName']
        assert isinstance(self.display_name, str)

        self.broadcaster_id = obj['broadcasterID']
        assert isinstance(self.broadcaster_id, str)

        self.source_location = obj['broadcasterLocation']
        assert isinstance(self.source_location, str)

        self.start_time = datetime.datetime.fromtimestamp(obj['startTime'], tz=pytz.utc)
        assert isinstance(self.start_time, datetime.datetime)

        self.preview_image_url = obj['previewImageURL']
        assert isinstance_or_none(self.preview_image_url, str)

        self.resizable_preview_image_url = obj['resizablePreviewImageURL']
        assert isinstance_or_none(self.resizable_preview_image_url, str)

        self.peak_listener_count = obj['peakListenerCount']
        assert isinstance(self.peak_listener_count, int)

        self.total_tune_in_count = obj['totalTuneInCount']
        assert isinstance(self.total_tune_in_count, int)

        self.media = MediaSet.parse(obj['media'])

    def __str__(self):  # pragma: no cover
        return f'<WebcastInfo {self.broadcaster_id} "{self.display_name}"">'


class RelativeBroadcasterLocation(Model):
    def __init__(self, obj: dict):
        super().__init__(obj)

        self.broadcaster = Broadcaster.parse(obj['broadcaster'])
        assert isinstance(self.broadcaster, Broadcaster)

        self.meters = obj['meters']
        assert isinstance(self.meters, int)

    def __str__(self):  # pragma: no cover
        return f'<RelativeBroadcasterLocation {self.broadcaster.broadcaster_id} - ~{round(self.meters/1000, 1)}km away>'


class SermonSeries(Model):
    def __init__(self, obj: dict):
        super().__init__(obj)

        self.type = obj.get('type')
        self.latest = dateutil.parser.parse(obj['latest']) if obj['latest'] is not None else None
        self.earliest = dateutil.parser.parse(obj['earliest']) if obj['earliest'] is not None else None
        self.title = obj['title']
        self.count = obj['count']

    def __str__(self):  # pragma: no cover
        return f'<SermonSeries {self.title} ({self.count} sermons)>'


class SermonEventType(enum.Enum):
    AUDIO_BOOK = 'Audio Book'
    BIBLE_STUDY = 'Bible Study'
    CAMP_MEETING = 'Camp Meeting'
    CHAPEL_SERVICE = 'Chapel Service'
    CHILDREN = 'Children'
    CLASSIC_AUDIO = 'Classic Audio'
    CONFERENCE = 'Conference'
    CURRENT_EVENTS = 'Current Events'
    DEBATE = 'Debate'
    DEVOTIONAL = 'Devotional'
    FUNERAL_SERVICE = 'Funeral Service'
    MIDWEEK_SERVICE = 'Midweek Service'
    PODCAST = 'Podcast'
    PRAYER_MEETING = 'Prayer Meeting'
    Q_AND_A = 'Question & Answer'
    RADIO_BROADCAST = 'Radio Broadcast'
    SPECIAL_MEETING = 'Special Meeting'
    SUNDAY_AFTERNOON = 'Sunday Afternoon'
    SUNDAY_AM = 'Sunday - AM'
    SUNDAY_PM = 'Sunday - PM'
    SUNDAY_SCHOOL = 'Sunday School'
    SUNDAY_SERVICE = 'Sunday Service'
    TEACHING = 'Teaching'
    TESTIMONY = 'Testimony'
    TV_BROADCAST = 'TV Broadcast'
    VIDEO_DVD = 'Video DVD'
    WEDDING = 'Wedding'
    YOUTH = 'Youth'


class SermonEventTypeDetail(Model):
    def __init__(self, obj: dict):
        super().__init__(obj)

        self.type = obj.get('type')
        self.description = obj.get('description')
        self.number_of_sermons = obj.get('numberOfSermons')
        self.roku_image = obj.get('rokuImage')
        self.roku_image_url = obj.get('rokuImageURL')
        self.fire_tv_image = obj.get('fireTVImage')
        self.fire_tv_image_url = obj.get('fireTVImageURL')


@enum.unique
class SpurgeonDevotionalType(enum.Enum):
    AM = 'AM'  # Morning Devotional
    PM = 'PM'  # Evening Devotional
    CHECKBOOK = 'CHECKBOOK'  # Faith's Checkbook (note American spelling)


class SpurgeonDevotional(Model):
    def __init__(self, obj: dict):
        super().__init__(obj)

        self.type = SpurgeonDevotionalType(obj['type'])
        self.month = obj['month']
        self.day = obj['day']
        self.quote = obj['quote']
        self.reference = obj['reference']
        self.content = obj['content']
        self.audio = Sermon(obj['sermon'])

    def __str__(self):  # pragma: no cover
        return f'<SpurgeonDevotional for {self.month}/{self.day}>'


@enum.unique
class OSISBook(enum.Enum):
    GEN = 'GEN'
    EXO = 'EXO'
    LEV = 'LEV'
    NUM = 'NUM'
    DEU = 'DEU'
    JOS = 'JOS'
    JDG = 'JDG'
    RUT = 'RUT'
    _1SA = '1SA'
    _2SA = '2SA'
    _1KI = '1KI'
    _2KI = '2KI'
    _1CH = '1CH'
    _2CH = '2CH'
    EZR = 'EZR'
    NEH = 'NEH'
    EST = 'EST'
    JOB = 'JOB'
    PSA = 'PSA'
    PRO = 'PRO'
    ECC = 'ECC'
    SNG = 'SNG'
    ISA = 'ISA'
    JER = 'JER'
    LAM = 'LAM'
    EZK = 'EZK'
    DAN = 'DAN'
    HOS = 'HOS'
    JOL = 'JOL'
    AMO = 'AMO'
    OBA = 'OBA'
    JON = 'JON'
    MIC = 'MIC'
    NAM = 'NAM'
    HAB = 'HAB'
    ZEP = 'ZEP'
    HAG = 'HAG'
    ZEC = 'ZEC'
    MAL = 'MAL'
    MAT = 'MAT'
    MRK = 'MRK'
    LUK = 'LUK'
    JHN = 'JHN'
    ACT = 'ACT'
    ROM = 'ROM'
    _1CO = '1CO'
    _2CO = '2CO'
    GAL = 'GAL'
    EPH = 'EPH'
    PHP = 'PHP'
    COL = 'COL'
    _1TH = '1TH'
    _2TH = '2TH'
    _1TI = '1TI'
    _2TI = '2TI'
    TIT = 'TIT'
    PHM = 'PHM'
    HEB = 'HEB'
    JAS = 'JAS'
    _1PE = '1PE'
    _2PE = '2PE'
    _1JN = '1JN'
    _2JN = '2JN'
    _3JN = '3JN'
    JUD = 'JUD'
    REV = 'REV'


@enum.unique
class SermonSortOption(enum.Enum):
    DOWNLOADS = 'downloads'
    EVENT = 'event'
    LANGUAGE = 'language'
    LAST_PLAYED = 'lastplayed'
    NEWEST = 'newest'
    OLDEST = 'oldest'
    PICK_DATE = 'pickdate'
    SERIES = 'series'
    SPEAKER = 'speaker'



import functools

from posixpath import join as posixjoin
from typing import Optional

from sermonaudio import API, APIException, models
from sermonaudio.utils import update_kwargs_for_key

URL_PATH = 'node'


class NodeAPIError(APIException):
    pass


def parse_node(response) -> models.Node:
    if response.ok:
        return models.Node.parse(response.json())
    else:
        raise NodeAPIError(response.json())


class Node(API):
    """An interface to the SermonAudio.com node API v2.

    This class provides a variety of low-level and high-level helpers to make it
    easy to interact with the API. A number of higher-level helpers are exposed to
    make common API calls easy. However, the API has many options, and we do not
    necessarily comprehensively model them all here. Please refer to the documentation,
    and if you need to build additional methods for your application, you may use the lower
    level helpers at the top of the class to build your own `Node` subclass with additional
    high-level helper methods.
    """
    #
    # Low-level building blocks
    #

    get_node = functools.partialmethod(API.get, parse_func=parse_node)
    get_sermon_list_node = functools.partialmethod(get_node,
                                                   parse_func=lambda res: [models.Sermon.parse(rec)
                                                                           for rec in parse_node(res).results])

    class PaginatedResponse:
        def __init__(self,
                     page: int,
                     page_size: int,
                     response: models.Node,
                     result_processing_func=None):
            self.page = page
            self.page_size = page_size
            self.next_url = response.next
            self.total_count = response.total_count
            self.results = [result_processing_func(result)
                            for result in response.results] if result_processing_func else response.results

    #
    # Sermons
    #

    @classmethod
    def get_sermons(cls,
                    path: str = None,
                    page: int = 1,
                    page_size: Optional[int] = None,
                    book: Optional[models.OSISBook] = None,
                    chapter: Optional[int] = None,
                    chapter_end: Optional[int] = None,
                    verse: Optional[int] = None,
                    verse_end: Optional[int] = None,
                    event_type: Optional[models.SermonEventType] = None,
                    search_keyword: Optional[str] = None,
                    language_code: Optional[str] = None,
                    require_audio: bool = False,
                    require_video: bool = False,
                    series: Optional[str] = None,
                    broadcaster_id: str = None,
                    speaker_name: str = None,
                    year: Optional[int] = None,
                    sort_by: models.SermonSortOption = models.SermonSortOption.NEWEST,
                    **kwargs) -> PaginatedResponse:
        """Returns a set of sermons with pagination data, when available.

        :param path: The path to the endpoint you want to call (defaults to the all-purpose sermon node endpoint).
        :param page: The page number to load (defaults to 1).
        :param page_size: The number of items per page (currently defaults to 50 if omitted).
        :param kwargs: Additional arguments to pass to the underlying get method.
        :param book: The book of the Bible to limit results to.
        :param chapter: The chapter of the book to limit results to.
        :param chapter_end: If you want to query a range of chapters, the range will span from chapter to chapter_end.
        :param verse: The verse of the chapter to limit results to.
        :param verse_end: If you want to query a range of verses, the range will span from verse to verse_end.
        :param event_type: The type of event to limit results to, if any.
        :param search_keyword: Keywords to search by.
        :param language_code: The language code to limit results to.
        :param require_audio: If true, all results will have ready-to-play audio.
        :param require_video: If true, all results will have ready-to-play video.
        :param series: The series to limit results to.
        :param broadcaster_id: The ID of the broadcaster to limit results to.
        :param speaker_name: The speaker name to limit results to.
        :param year: The year to limit results to (preach date, not upload date).
        :param sort_by: A sort to apply to the results. Defaults to preach date descending.
        :return: A sermon list response with pagination data, if available.
        """
        path = path or posixjoin(URL_PATH, 'sermons')
        update_kwargs_for_key(kwargs, 'params', {
            'page': page,
            'pageSize': page_size,
            'book': book.value if book else None,
            'chapter': chapter,
            'chapterEnd': chapter_end,
            'verse': verse,
            'verseEnd': verse_end,
            'eventType': event_type.value if event_type else None,
            'searchKeyword': search_keyword,
            'languageCode': language_code,
            'requireAudio': require_audio,
            'requireVideo': require_video,
            'series': series,
            'broadcasterID': broadcaster_id,
            'speakerName': speaker_name,
            'year': year,
            'sortBy': sort_by.value
        })

        return cls.get_node(path,
                            parse_func=lambda res: cls.PaginatedResponse(page,
                                                                         page_size,
                                                                         parse_node(res),
                                                                         models.Sermon.parse),
                            **kwargs)

    @classmethod
    def get_sermon(cls, sermon_id: str, **kwargs) -> Optional[models.Sermon]:
        """Fetches info for a single sermon, if it exists.

        :param sermon_id: The ID of the sermon you want to fetch.
        :param kwargs: Additional arguments to pass to the underlying get method.
        :return: types.Sermon or None
        """
        return cls.get(posixjoin(URL_PATH, 'sermons', str(sermon_id)),
                       parse_func=lambda res: models.Sermon.parse(res.json()) if res.ok else None,
                       **kwargs)

    #
    # Broadcasters
    #

    @classmethod
    def get_broadcaster(cls, broadcaster_id: str, **kwargs) -> Optional[models.Broadcaster]:
        """Fetches info for a single broadcaster, if it exists.

        :param broadcaster_id: The ID of the broadcaster you want to fetch.
        :param kwargs: Additional arguments to pass to the underlying get method.
        :return: types.Broadcaster or None
        """
        return cls.get(posixjoin(URL_PATH, 'broadcasters', str(broadcaster_id)),
                       parse_func=lambda res: models.Broadcaster.parse(res.json()) if res.ok else None,
                       **kwargs)

    @classmethod
    def get_broadcasters_near_location(cls,
                                       latitude: float,
                                       longitude: float,
                                       meters: int,
                                       **kwargs) -> [models.RelativeBroadcasterLocation]:
        """Fetches a list of broadcasters near a location.

        :param latitude: The latitude of the search origin.
        :param longitude: The longitude of the search origin.
        :param meters: The distance from the origin to search.
        :param kwargs: Additional arguments to pass to the underlying get method.
        :return: A list of broadcaster relative location objects
        """
        update_kwargs_for_key(kwargs, 'params', {
            'latitude': latitude,
            'longitude': longitude,
            'meters': meters
        })
        return cls.get(posixjoin(URL_PATH, 'broadcasters', 'near_location'),
                       parse_func=lambda res: [models.RelativeBroadcasterLocation.parse(rec)
                                               for rec in parse_node(res).results],
                       **kwargs)

    @classmethod
    def get_series_list(cls,
                        broadcaster_id: str = None,
                        **kwargs) -> [models.SermonSeries]:
        """Fetches a sermon series list for a broadcaster.

        :param broadcaster_id: You can omit this if you are using a broadcaster API key.
        :param kwargs: Additional arguments to pass to the underlying get method.
        :return: A list of series objects
        """
        return cls.get(posixjoin(URL_PATH, 'broadcasters', broadcaster_id, 'filter_options', 'series'),
                       parse_func=lambda res: [models.SermonSeries.parse(rec) for rec in parse_node(res).results],
                       **kwargs)

    @classmethod
    def get_sermon_event_types(cls,
                               broadcaster_id: str = None,
                               **kwargs) -> [models.SermonEventTypeDetail]:
        """Fetches a sermon series list for a broadcaster.

        :param broadcaster_id: You can omit this if you are using a broadcaster API key.
        :param kwargs: Additional arguments to pass to the underlying get method.
        :return: A list of sermon event type detail objects
        """
        update_kwargs_for_key(kwargs, 'params', {
            'broadcaster_id': broadcaster_id
        })
        return cls.get(posixjoin(URL_PATH, 'filter_options', 'sermon_event_types'),
                       parse_func=lambda res: [models.SermonEventTypeDetail.parse(rec)
                                               for rec in parse_node(res).results],
                       **kwargs)

    #
    # Misc
    #

    @classmethod
    def get_webcasts_in_progress(cls,
                                 broadcaster_id: str = None,
                                 **kwargs) -> [models.Webcast]:
        """Fetches a list of webcasts that are currently in progress.

        :param broadcaster_id: You can omit this if you are using a broadcaster API key.
        :param kwargs: Additional arguments to pass to the underlying get method.
        :return: A list of webcasts that are live now
        """
        update_kwargs_for_key(kwargs, 'params', {
            'sourceID': broadcaster_id
        })
        return cls.get(posixjoin(URL_PATH, 'webcasts'),
                       parse_func=lambda res: [models.Webcast.parse(rec) for rec in parse_node(res).results],
                       **kwargs)

    @classmethod
    def get_spurgeon_devotional(cls,
                                kind: models.SpurgeonDevotionalType,
                                month: int,
                                day_of_month: int,
                                **kwargs) -> Optional[models.SpurgeonDevotional]:
        """Returns the Spurgeon devotional for a given month and day

        :param kind: The kind of devotional (one of the values in types.SpurgeonDevotionalType).
        :param month: The month component of the date.
        :param day_of_month: The day of month component of the date.
        :param kwargs: Additional arguments to pass to the underlying get method.
        :return: A list devotional object or None
        """
        update_kwargs_for_key(kwargs, 'params', {
            'type': kind.value,
            'month': month,
            'day_of_month': day_of_month
        })

        def parse_spurgeon(result):
            data = parse_node(result).results

            return models.SpurgeonDevotional.parse(data) if data else None

        return cls.get(posixjoin(URL_PATH, 'spurgeon_devotional'),
                       parse_func=parse_spurgeon,
                       **kwargs)



from setuptools import setup

with open('README.md') as fp:
    readme = fp.read()

setup(name='sermonaudio',
      version='2.0.1',
      description='The official Python client library for accessing the SermonAudio.com APIs',
      long_description=readme,
      long_description_content_type='text/markdown',
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3.6',
      ],
      url='http://api.sermonaudio.com/',
      author='SermonAudio.com',
      author_email='info@sermonaudio.com',
      keywords='sermon audio sermonaudio API preaching church bible',
      license='MIT',
      packages=['sermonaudio', 'sermonaudio.node', 'sermonaudio.broadcaster'],
      install_requires=[
          'requests>=2.19.1',
          'pytz>=2018.5',
          'python-dateutil>=2.7.3'
      ])
